#!/bin/bash

# crc_diagram.sh draws circuit diagrams for a CRC hexadecimal
# polynomial.
# Copyright (C) 2022 Remington Furman
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

usage () {
    cat <<EOF
Usage: $1 (-p powers | -h poly) [args]
  Explicit representation:
    -p powers  Use comma separated list of powers to set polynomial,
               ignore hexadecimal related arguments

  Hexadecimal representation:
    -h poly    Hexadecimal representation of polynomial
    Representation presets:
    -f         Forward/normal     polynomial format
    -r         Reverse            polynomial format
    -k         Koopman            polynomial format
    -c         reCiprocal         polynomial format
    -v         reVerse reciprocal polynomial format
    Representation parameters:
    -a         Ascending  polynomial (MSB is x^0) (default)
    -d         Descending polynomial (MSB is x^n)
    -n degree  Polynomial degree (x^N)
    -o         One   term (x^0) is implicit (default: explicit)
    -m         Monic term (x^n) is implicit (default: explicit)

  Diagram options:
    -l         Left shift the diagram (default: right)

  Miscellaneous:
    -u         Usage (help)
EOF
}


# Helper functions
dec () {
    # Convert number to decimal.
    printf "%d" "$1"
}
bin () {
    # Convert number to binary.
    bc <<<"obase=2;$(dec $1)"
}

superscript () {
    if true ; then
        # Replace digits with Unicode superscript characters.
        sed 'y/0123456789/⁰¹²³⁴⁵⁶⁷⁸⁹/' <<< "$1"
    else
        # Prefix digits with ^.
        echo "^${1}"
    fi
}


# Set default parameters.
POLY_HEX=
IMPLICIT_ONE=    # Blank if the term is already included in the hex poly,
IMPLICIT_MONIC=  # otherwise 1.
ASCENDING=true   # MSB of hex constant holds x^0.
                 # (It's ascending left to right when printed in binary.)
DEGREE=          # Necessary when degree can't be inferred from format.
FWD="right"      # Plot shift direction.
REV="left"       # Opposite of shift direction.


# Read command line arguments.
optstring="p:h:frkcvadn:omlu"
while getopts ${optstring} arg; do
    case ${arg} in
        p)                      # Explicit polynomial representation
            declare -i POLY_INT=0
            IFS=',' read -ra POWERS <<< "${OPTARG}"
            for i in "${POWERS[@]}"; do
                ((POLY_INT |= 1<<"$i"))
            done
            POLY_BIN=$(bin "$POLY_INT" | rev)
            ;;
        h)
            POLY_HEX="${OPTARG}"
            ;;
        f)                      # Forward / Normal preset
            IMPLICIT_ONE=
            IMPLICIT_MONIC=1
            ASCENDING=false
            ;;
        r)                      # Reverse preset
            IMPLICIT_ONE=
            IMPLICIT_MONIC=1
            ASCENDING=true
            ;;
        k)                      # Koopman preset
            IMPLICIT_ONE=1
            IMPLICIT_MONIC=
            ASCENDING=false
            ;;
        c)                      # Reciprocal preset
            IMPLICIT_ONE=
            IMPLICIT_MONIC=1
            ASCENDING=false
            ;;
        v)                      # Reverse reciprocal preset
            IMPLICIT_ONE=
            IMPLICIT_MONIC=1
            ASCENDING=true
            ;;
        a)
            ASCENDING=true
            ;;
        d)
            ASCENDING=false
            ;;
        n)
            # Use this to determine length of polynomial when it can't
            # be inferred:
            #   Ascending  with implicit x^0, or
            #   Descending with implicit x^n.
            DEGREE="${OPTARG}"
            ;;
        o)
            IMPLICIT_ONE=1
            ;;
        m)
            IMPLICIT_MONIC=1
            ;;
        l)
            FWD="left"
            REV="right"
            ;;
        u)
            usage $0
            exit
            ;;
    esac
done


# Determine the explicit ascending binary representation of the
# polynomial from the command line arguments.
if [ -n "$POLY_BIN" ] ; then
    # We already have the explicit representation.
    POLY_HEX=
else
    # Convert the hexadecimal representation into an explicit
    # ascending binary representation.

    declare -i POLY_INT="${POLY_HEX}"
    if   "$ASCENDING" && [ -n "$IMPLICIT_ONE" ] ||
       ! "$ASCENDING" && [ -n "$IMPLICIT_MONIC" ]; then
        # Can't infer polynomial degree, so we need it from the user.

        if [ -z "$DEGREE" ] ; then
            >&2 echo "Can't infer polynomial degree. Please provide '-n degree' flag."
            exit 1
        else
            # Use DEGREE to add all implied bits to the polynomial.
            if "$ASCENDING" ; then
                if [ -n "$IMPLICIT_MONIC" ] ; then
                    ((POLY_INT <<= 1))
                    ((POLY_INT  |= 1))
                    IMPLICIT_MONIC=
                fi
                IMPLICIT_ONE=   # This term will be explicit soon.
            else
                if [ -n "$IMPLICIT_ONE" ] ; then
                    ((POLY_INT <<= 1))
                    ((POLY_INT  |= 1))
                    IMPLICIT_MONIC=
                fi
                IMPLICIT_MONIC= # This term will be explicit soon.
            fi
            ((POLY_INT |= 1<<"$DEGREE"))
        fi
    fi

    POLY_BIN=$(bin "${POLY_INT}")

    if ! $ASCENDING ; then
        POLY_BIN=$(echo "${POLY_BIN}" | rev)
    fi
    POLY_BIN="${IMPLICIT_ONE}${POLY_BIN}${IMPLICIT_MONIC}"

fi

# Make sure we have a polynomial of at least degree 1.
if (("${#POLY_BIN}" < 2)) ; then
    >&2 echo "Error: Supplied polynomial too short."
    exit 1
fi


# Pretty print polynomial.
POLY="x$(superscript 0)"  # or POLY="1", if you prefer
for i in $(seq 1 $(("${#POLY_BIN}" - 1))) ; do
    BIT="${POLY_BIN:$i:1}"
    if [ "$BIT" -eq 1 ] ; then
        POLY+=" + x$(superscript $i)"
    fi
done
echo "# ${POLY_HEX:+$POLY_HEX -> }$POLY_BIN -> $POLY"

# bit, term
flipflop () {
    echo "arrow \"\" \"$2\" big big"
    echo "box \"$1\" big big"
}

# Pikchr label
tap () {
    echo "arrow"
    echo "$1: circle \"+\" big big"
}

# Generate Pikchr diagram.
echo "boxwid = 0.5in; boxht = 0.5in;"

# Shift register and XOR taps
echo "# Shift register and XOR taps"
echo "$FWD"
echo -n "Start: "
flipflop "1" "x$(superscript 0)"

POLY_DEGREE=$(("${#POLY_BIN}" - 1))
for i in $(seq 1 $(("$POLY_DEGREE" - 1))) ; do
    BIT="${POLY_BIN:$i:1}"
    if [ "$BIT" -eq 1 ] ; then
        TAP_LABEL="Tap${i}"
        TAPS+="$TAP_LABEL "
        tap "$TAP_LABEL"
    fi
    flipflop "$BIT" "x$(superscript $i)"
done

# Input
echo "# Input"
echo "arrow \"\" \"x$(superscript ${POLY_DEGREE})\" big big"
echo "Input: circle \"+\" big big"
echo "arrow <- from Input.s down"
echo "line $REV \"Input\" big big \"\""

# Feedback network
echo "# Feedback network"
echo "FB: line from Input.n up \\"
echo "    then $REV until even with Start.start \\"
echo "    then to Start.start"

for TAP in ${TAPS} ; do
    echo "arrow <- from ${TAP}.n up until even with FB.n; dot"
done
